/*
    Drools5 Integration Helper
    Copyright (C) 2009  Mathieu Boretti mathieu.boretti@gmail.com

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

 */
package org.boretti.drools.integration.drools5.implementation;

import java.lang.annotation.Annotation;
import java.lang.reflect.Proxy;

import org.apache.log4j.Logger;
import org.boretti.drools.integration.drools5.DroolsInterface;
import org.boretti.drools.integration.drools5.DroolsServiceType;
import org.boretti.drools.integration.drools5.annotations.DroolsService;
import org.boretti.drools.integration.drools5.exceptions.NotAnInterfaceDroolsError;
import org.boretti.drools.integration.drools5.exceptions.NotAnnotatedClassDroolsError;
import org.boretti.drools.integration.drools5.exceptions.NotAnnotatedInterfaceDroolsError;
import org.boretti.drools.integration.drools5.implementation.DroolsInterfaceImplementation;
import org.boretti.drools.integration.drools5.implementation.DroolsCompiledInterfaceImplementation;
import org.drools.RuleBase;

public class DroolsProviderImplementation {
	
	private final Logger logger = Logger.getLogger(this.getClass());
	
	private DroolsProviderImplementation() {}
	
	private static final DroolsProviderImplementation instance = new DroolsProviderImplementation();
	
	public static final DroolsProviderImplementation getInstance() {
		return instance;
	}
	
	@SuppressWarnings("unchecked")
	public <T> T getService(Class<T> clazz) {
		if (logger.isDebugEnabled()) logger.debug("Gettting service for "+clazz);
		validNullability(clazz,"Invalid interface - can't be null");
		if (!clazz.isInterface()) {
			String msg = "Not an interface "+clazz;
			logger.error(msg);
			throw new NotAnInterfaceDroolsError(msg);
		}
		DroolsService ds = getAnnotation(clazz);
		if(ds==null) {
			String msg = "Invalid interface (missing annotation) "+clazz;
			logger.error(msg);
			throw new NotAnnotatedInterfaceDroolsError(msg);
		}
		DroolsServiceType dst=ds.type();
		String resource = ds.resourceName();
		validNullability(dst,"Invalid type parameter of annotation - can't be null");
		DroolsInterfaceImplementation<T> inter=getDroolsImplementation(dst,resource,clazz);
		if (logger.isDebugEnabled()) logger.debug("Creating proxy using "+clazz+" from "+dst+" using resource "+resource);
		return (T) Proxy.newProxyInstance(clazz.getClassLoader(),new Class[] {clazz,DroolsInterface.class},new DroolsProxy<T>(clazz,inter));
	}
	
	/**
	 * 
	 * @param clazz
	 * @return
	 * @since 1.0.0
	 */
	public <T> RuleBase getRuleBase(Class<T> clazz) {
		if (logger.isDebugEnabled()) logger.debug("Gettting service for "+clazz);
		validNullability(clazz,"Invalid class or interface - can't be null");
		DroolsService ds = getAnnotation(clazz);
		if(ds==null) {
			if (clazz.isInterface()) {
				String msg = "Invalid interface (missing annotation) "+clazz;
				logger.error(msg);
				throw new NotAnnotatedInterfaceDroolsError(msg);
			} else {
				String msg = "Invalid class (missing annotation) "+clazz;
				logger.error(msg);
				throw new NotAnnotatedClassDroolsError(msg);
			}
		}
		DroolsServiceType dst=ds.type();
		String resource = ds.resourceName();
		validNullability(dst,"Invalid type parameter of annotation - can't be null");
		return getDroolsImplementation(dst,resource,clazz).getRuleBase();
	}
	
	private <T> DroolsInterfaceImplementation<T> getDroolsImplementation(DroolsServiceType dst,String resource,Class<T> clazz) {
		if (DroolsServiceType.COMPILED.equals(dst)) {
			if (resource==null || resource.trim().equalsIgnoreCase("")) resource = clazz.getName()+".cdrl";
			return new DroolsCompiledInterfaceImplementation<T>(clazz, resource);
		} else if (DroolsServiceType.SOURCE.equals(dst)){
			if (resource==null || resource.trim().equalsIgnoreCase("")) resource = clazz.getName()+".drl";
			return new DroolsSourceInterfaceImplementation<T>(clazz,resource);
		} else {
			if (resource==null || resource.trim().equalsIgnoreCase("")) resource = clazz.getName()+".xml";
			return new DroolsXMLInterfaceImplementation<T>(clazz,resource);
		}
	}
	
	private void validNullability(Object o,String msg) {
		if (o==null) {
			logger.error(msg);
			throw new IllegalArgumentException(msg);
		}
	}
	
	private <T> DroolsService getAnnotation(Class<T> clazz) {
		Annotation as[] = clazz.getAnnotations();
		for(Annotation a:as) {
			if (a.annotationType().equals(DroolsService.class)) return (DroolsService)a;
		}
		return null;
	}
}
